"""
Objects that act like dict, list, OrderedDict but allow
lazy conversion of tagged ASDF tree nodes to custom objects.
"""

import collections
import inspect
import warnings
import weakref

from . import tagged, treeutil, yamlutil
from .config import get_config
from .exceptions import AsdfConversionWarning, AsdfLazyReferenceError
from .extension._serialization_context import BlockAccess

__all__ = ["AsdfDictNode", "AsdfListNode", "AsdfOrderedDictNode"]


class _TaggedObjectCacheItem:
    """
    A tagged node and a (weakref) to the converted custom object
    """

    def __init__(self, tagged_node, custom_object):
        self.tagged_node = tagged_node
        try:
            self._custom_object_ref = weakref.ref(custom_object)
        except TypeError:
            # if a weakref is not possible, store the object
            self._custom_object_ref = lambda obj=custom_object: obj

    @property
    def custom_object(self):
        return self._custom_object_ref()


class _TaggedObjectCache:
    """
    A cache of tagged nodes and their corresponding custom objects.

    This is critical for trees that contain references/pointers to the
    same object at multiple locations in the tree.

    Only weakrefs are key to the custom objects to allow large items
    deleted from the tree to be garbage collected. This means that an
    item added to the cache may later fail to retrieve (if the weakref-ed
    custom object was deleted).
    """

    def __init__(self):
        # start with a clear cache
        self.clear()

    def clear(self):
        self._cache = {}

    def retrieve(self, tagged_node):
        """
        Check the cache for a previously converted object.

        Parameters
        ----------
        tagged_node : Tagged
            The tagged representation of the custom object

        Returns
        -------
        custom_object : None or the converted object
            The custom object previously converted from the tagged_node or
            ``None`` if the object hasn't been converted (or was previously
            deleted from the tree).
        """
        key = id(tagged_node)
        if key not in self._cache:
            return None
        item = self._cache[key]
        custom_object = item.custom_object
        if custom_object is None:
            del self._cache[key]
        return custom_object

    def store(self, tagged_node, custom_object):
        """
        Store a converted custom object in the cache.

        Parameters
        ----------
        tagged_node : Tagged
            The tagged representation of the custom object

        custom_object : converted object
            The custom object (a weakref to this object will be kept in the cache).
        """
        self._cache[id(tagged_node)] = _TaggedObjectCacheItem(tagged_node, custom_object)


def _resolve_af_ref(af_ref):
    msg = "Failed to resolve AsdfFile reference"
    if af_ref is None:
        raise AsdfLazyReferenceError(msg)
    af = af_ref()
    if af is None:
        raise AsdfLazyReferenceError(msg)
    return af


def _to_lazy_node(node, af_ref):
    """
    Convert an object to a _AsdfNode subclass.
    If the object does not have a corresponding subclass
    it will be returned unchanged.
    """
    if isinstance(node, tagged.TaggedList):
        return tagged.TaggedList(data=_to_lazy_node(node.data, af_ref), tag=node._tag)
    elif isinstance(node, tagged.TaggedDict):
        return tagged.TaggedDict(data=_to_lazy_node(node.data, af_ref), tag=node._tag)
    elif isinstance(node, list):
        return AsdfListNode(node, af_ref)
    elif isinstance(node, collections.OrderedDict):
        return AsdfOrderedDictNode(node, af_ref)
    elif isinstance(node, dict):
        return AsdfDictNode(node, af_ref)
    return node


class _AsdfNode:
    """
    The "lazy node" base class that handles object
    conversion and wrapping and contains a weak reference
    to the `asdf.AsdfFile` that triggered the creation of this
    node (when the "lazy tree" was loaded).
    """

    def __init__(self, data=None, af_ref=None):
        self._af_ref = af_ref
        self.data = data

    @property
    def tagged(self):
        """
        Return the tagged tree backing this node
        """
        return self.data

    def __deepcopy__(self, memo):
        return treeutil.walk_and_modify(self, lambda n: n)

    def _convert_and_cache(self, value, key):
        """
        Convert ``value`` to either:

            - a custom object if ``value`` is `asdf.tagged.Tagged`
            - an ``asdf.lazy_nodes.AsdfListNode` if ``value`` is
              a ``list``
            - an ``asdf.lazy_nodes.AsdfDictNode` if ``value`` is
              a ``dict``
            - an ``asdf.lazy_nodes.AsdfOrderedDictNode` if ``value`` is
              a ``OrderedDict``
            - otherwise return ``value`` unmodified

        After conversion the result (``obj``) will be stored in this
        `asdf.lazy_nodes._AsdfNode` using the provided key and cached
        in the corresponding `asdf.AsdfFile` instance (so other
        references to ``value`` in the tree will return the same
        ``obj``).

        Parameters
        ----------
        value :
            The object to convert from a Tagged to custom object
            or wrap with an _AsdfNode or return unmodified.

        key :
            The key under which the converted/wrapped object will
            be stored.


        Returns
        -------
        obj :
            The converted or wrapped (or the value if no conversion
            or wrapping is required).
        """
        # if the value has already been wrapped, return it
        if isinstance(value, _AsdfNode):
            return value
        if isinstance(value, (tagged.TaggedDict, tagged.TaggedList)) and isinstance(value.data, _AsdfNode):
            return value
        if not isinstance(value, tagged.Tagged) and type(value) not in _base_type_to_node_map:
            return value
        af = _resolve_af_ref(self._af_ref)
        # if the obj that will be returned from this value
        # is already cached, use the cached obj
        if (obj := af._tagged_object_cache.retrieve(value)) is not None:
            self[key] = obj
            return obj
        # for Tagged instances, convert them to their custom obj
        if isinstance(value, tagged.Tagged):
            extension_manager = af.extension_manager
            tag = value._tag
            if not extension_manager.handles_tag(tag):
                if not af._ignore_unrecognized_tag:
                    warnings.warn(
                        f"{tag} is not recognized, converting to raw Python data structure",
                        AsdfConversionWarning,
                    )
                obj = _to_lazy_node(value, self._af_ref)
            else:
                converter = extension_manager.get_converter_for_tag(tag)
                if not getattr(converter, "lazy", False) or inspect.isgeneratorfunction(
                    converter._delegate.from_yaml_tree
                ):
                    obj = yamlutil.tagged_tree_to_custom_tree(value, af)
                else:
                    data = _to_lazy_node(value.data, self._af_ref)
                    sctx = af._create_serialization_context(BlockAccess.READ)
                    try:
                        obj = converter.from_yaml_tree(data, tag, sctx)
                    except Exception as err:
                        if get_config().warn_on_failed_conversion:
                            warnings.warn(f"A node failed to convert with: {err}", AsdfConversionWarning)
                            obj = _to_lazy_node(value, self._af_ref)
                        else:
                            raise
                    sctx.assign_object(obj)
                    sctx.assign_blocks()
                    sctx._mark_extension_used(converter.extension)
        else:
            # for non-tagged objects, wrap in an _AsdfNode
            node_type = _base_type_to_node_map[type(value)]
            obj = node_type(value, self._af_ref)
        # cache the converted/wrapped obj with the AsdfFile so other
        # references to the same Tagged value will result in the
        # same obj
        af._tagged_object_cache.store(value, obj)
        self[key] = obj
        return obj


class AsdfListNode(_AsdfNode, collections.UserList):
    """
    An class that acts like a ``list``. The items in this ``list``
    will start out as tagged nodes which will only be converted to
    custom objects the first time they are indexed (the custom object
    will then be cached for later reuse).

    If sliced, this will return a new instance of `AsdfListNode` for
    the sliced portion of the list.
    """

    def __init__(self, data=None, af_ref=None):
        if data is None:
            data = []
        _AsdfNode.__init__(self, data, af_ref)
        collections.UserList.__init__(self, data)

    def __copy__(self):
        return AsdfListNode(self.data.copy(), self._af_ref)

    def __eq__(self, other):
        if self is other:
            return True
        return list(self) == list(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getitem__(self, key):
        # key might be an int or slice
        value = super().__getitem__(key)
        if isinstance(key, slice):
            return AsdfListNode(value, self._af_ref)
        return self._convert_and_cache(value, key)


class AsdfDictNode(_AsdfNode, collections.UserDict):
    """
    An class that acts like a ``dict``. The values for this
    ``dict`` will start out as tagged nodes which will only
    be converted to custom objects the first time the corresponding
    key is used (the custom object will then be cached for later
    reuse).
    """

    def __init__(self, data=None, af_ref=None):
        if data is None:
            data = {}
        _AsdfNode.__init__(self, data, af_ref)
        collections.UserDict.__init__(self, data)

    def __copy__(self):
        return AsdfDictNode(self.data.copy(), self._af_ref)

    def __eq__(self, other):
        if self is other:
            return True
        return dict(self) == dict(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getitem__(self, key):
        return self._convert_and_cache(super().__getitem__(key), key)


class AsdfOrderedDictNode(AsdfDictNode, collections.OrderedDict):
    """
    An class that acts like a ``collections.OrderedDict``. The values
    for this ``OrderedDict`` will start out as tagged nodes which will only
    be converted to custom objects the first time the corresponding
    key is used (the custom object will then be cached for later
    reuse).
    """

    def __init__(self, data=None, af_ref=None):
        if data is None:
            data = collections.OrderedDict()
        AsdfDictNode.__init__(self, data, af_ref)

    def __copy__(self):
        return AsdfOrderedDictNode(self.data.copy(), self._af_ref)


_base_type_to_node_map = {
    dict: AsdfDictNode,
    list: AsdfListNode,
    collections.OrderedDict: AsdfOrderedDictNode,
}



import os
import warnings

import pytest

# We ignore these files because these modules create deprecation warnings on
# import. When warnings are turned into errors this will completely prevent
# test collection
collect_ignore = ["asdf.py", "stream.py"]


def pytest_collection_modifyitems(items):
    # first check if warnings are already turned into errors
    for wf in warnings.filters:
        if wf == ("error", None, Warning, None, 0):
            return
    # Turn warnings into errors for all tests, this is needed
    # as running tests through pyargs will not use settings
    # defined in pyproject.toml
    for item in items:
        item.add_marker(pytest.mark.filterwarnings("error"), False)


@pytest.fixture(scope="session", autouse=True)
def _temp_cwd(tmp_path_factory):
    """
    This fixture creates a temporary current working directory
    for the test session, so that docstring tests that write files
    don't clutter up the real cwd.
    """
    original_cwd = os.getcwd()
    try:
        os.chdir(tmp_path_factory.mktemp("cwd"))
        yield
    finally:
        os.chdir(original_cwd)


def pytest_addoption(parser):
    parser.addoption(
        "--jsonschema",
        action="store_true",
        default=False,
        help="Run jsonschema test suite tests",
    )



import threading

import asdf_standard.integration
import pytest

import asdf
from asdf import get_config
from asdf._core._integration import get_json_schema_resource_mappings
from asdf.extension import ExtensionProxy
from asdf.resource import ResourceMappingProxy


def test_config_context():
    assert get_config().validate_on_read is True

    with asdf.config_context() as config:
        config.validate_on_read = False
        assert get_config().validate_on_read is False

    assert get_config().validate_on_read is True


def test_config_context_nested():
    assert get_config().validate_on_read is True

    with asdf.config_context() as config1:
        config1.validate_on_read = False
        with asdf.config_context() as config2:
            config2.validate_on_read = True
            with asdf.config_context() as config3:
                config3.validate_on_read = False
                assert get_config().validate_on_read is False

    assert get_config().validate_on_read is True


def test_config_context_threaded():
    assert get_config().validate_on_read is True

    thread_value = None

    def worker():
        nonlocal thread_value
        thread_value = get_config().validate_on_read
        with asdf.config_context() as config:
            config.validate_on_read = False

    with asdf.config_context() as config:
        config.validate_on_read = False
        thread = threading.Thread(target=worker)
        thread.start()
        thread.join()

    assert thread_value is True
    assert get_config().validate_on_read is True


def test_global_config():
    assert get_config().validate_on_read is True

    get_config().validate_on_read = False
    assert get_config().validate_on_read is False

    with asdf.config_context() as config:
        assert config.validate_on_read is False
        config.validate_on_read = True
        assert get_config().validate_on_read is True

    assert get_config().validate_on_read is False

    # Global config is reset to defaults by autouse
    # fixture in asdf/_tests/conftest.py.


def test_validate_on_read():
    with asdf.config_context() as config:
        assert config.validate_on_read == asdf.config.DEFAULT_VALIDATE_ON_READ
        config.validate_on_read = False
        assert get_config().validate_on_read is False
        config.validate_on_read = True
        assert get_config().validate_on_read is True


def test_default_version():
    with asdf.config_context() as config:
        assert config.default_version == asdf.config.DEFAULT_DEFAULT_VERSION
        assert asdf.config.DEFAULT_DEFAULT_VERSION != "1.2.0"
        config.default_version = "1.2.0"
        assert config.default_version == "1.2.0"
        with pytest.raises(ValueError, match=r"ASDF Standard version .* is not supported by asdf==.*"):
            config.default_version = "0.1.5"


def test_legacy_fill_schema_defaults():
    with asdf.config_context() as config:
        assert config.legacy_fill_schema_defaults == asdf.config.DEFAULT_LEGACY_FILL_SCHEMA_DEFAULTS
        config.legacy_fill_schema_defaults = False
        assert get_config().legacy_fill_schema_defaults is False
        config.legacy_fill_schema_defaults = True
        assert get_config().legacy_fill_schema_defaults is True


def test_array_inline_threshold():
    with asdf.config_context() as config:
        assert config.array_inline_threshold == asdf.config.DEFAULT_ARRAY_INLINE_THRESHOLD
        config.array_inline_threshold = 10
        assert get_config().array_inline_threshold == 10
        config.array_inline_threshold = None
        assert get_config().array_inline_threshold is None


def test_all_array_storage():
    with asdf.config_context() as config:
        assert config.all_array_storage == asdf.config.DEFAULT_ALL_ARRAY_STORAGE
        config.all_array_storage = "internal"
        assert get_config().all_array_storage == "internal"
        config.all_array_storage = None
        assert get_config().all_array_storage is None
        with pytest.raises(ValueError, match=r"Invalid value for all_array_storage"):
            config.all_array_storage = "foo"


def test_all_array_compression():
    with asdf.config_context() as config:
        assert config.all_array_compression == asdf.config.DEFAULT_ALL_ARRAY_COMPRESSION
        config.all_array_compression = "zlib"
        assert get_config().all_array_compression == "zlib"
        config.all_array_compression = None
        assert get_config().all_array_compression is None
        with pytest.raises(ValueError, match=r"Supported compression types are"):
            config.all_array_compression = "foo"


def test_all_array_compression_kwargs():
    with asdf.config_context() as config:
        assert config.all_array_compression_kwargs == asdf.config.DEFAULT_ALL_ARRAY_COMPRESSION_KWARGS
        config.all_array_compression_kwargs = {}
        assert get_config().all_array_compression_kwargs == {}
        config.all_array_compression_kwargs = None
        assert get_config().all_array_compression_kwargs is None
        with pytest.raises(ValueError, match=r"Invalid value for all_array_compression_kwargs"):
            config.all_array_compression_kwargs = "foo"


def test_resource_mappings():
    with asdf.config_context() as config:
        core_mappings = get_json_schema_resource_mappings() + asdf_standard.integration.get_resource_mappings()

        default_mappings = config.resource_mappings
        assert len(default_mappings) >= len(core_mappings)

        new_mapping = {"http://somewhere.org/schemas/foo-1.0.0": b"foo"}
        config.add_resource_mapping(new_mapping)

        assert len(config.resource_mappings) == len(default_mappings) + 1
        assert any(m for m in config.resource_mappings if m.delegate is new_mapping)

        # Adding a mapping should be idempotent:
        config.add_resource_mapping(new_mapping)
        # ... even if wrapped:
        config.add_resource_mapping(ResourceMappingProxy(new_mapping))
        assert len(config.resource_mappings) == len(default_mappings) + 1

        # Adding a mapping should place it at the front of the line:
        front_mapping = {"http://somewhere.org/schemas/baz-1.0.0": b"baz"}
        config.add_resource_mapping(front_mapping)
        assert len(config.resource_mappings) == len(default_mappings) + 2
        assert config.resource_mappings[0].delegate is front_mapping

        # ... even if the mapping is already in the list:
        config.add_resource_mapping(new_mapping)
        assert len(config.resource_mappings) == len(default_mappings) + 2
        assert config.resource_mappings[0].delegate is new_mapping

        # Reset should get rid of any additions:
        config.reset_resources()
        assert len(config.resource_mappings) == len(default_mappings)

        # Should be able to remove a mapping:
        config.add_resource_mapping(new_mapping)
        config.remove_resource_mapping(new_mapping)
        assert len(config.resource_mappings) == len(default_mappings)

        # ... even if wrapped:
        config.add_resource_mapping(new_mapping)
        config.remove_resource_mapping(ResourceMappingProxy(new_mapping))
        assert len(config.resource_mappings) == len(default_mappings)

        # ... and also by the name of the package the mappings came from:
        config.add_resource_mapping(ResourceMappingProxy(new_mapping, package_name="foo"))
        config.add_resource_mapping(
            ResourceMappingProxy({"http://somewhere.org/schemas/bar-1.0.0": b"bar"}, package_name="foo"),
        )
        config.remove_resource_mapping(package="foo")
        assert len(config.resource_mappings) == len(default_mappings)

        # Can combine the package and mapping filters when removing:
        config.add_resource_mapping(ResourceMappingProxy(new_mapping, package_name="foo"))
        config.remove_resource_mapping(new_mapping, package="foo")
        assert len(config.resource_mappings) == len(default_mappings)

        # But not omit both:
        with pytest.raises(ValueError, match=r"Must specify at least one of mapping or package"):
            config.remove_resource_mapping()

        # Removing a mapping should be idempotent:
        config.add_resource_mapping(new_mapping)
        config.remove_resource_mapping(new_mapping)
        config.remove_resource_mapping(new_mapping)
        assert len(config.resource_mappings) == len(default_mappings)


def test_resource_manager():
    with asdf.config_context() as config:
        # Initial resource manager should contain just the entry points resources:
        assert "http://stsci.edu/schemas/asdf/core/asdf-1.1.0" in config.resource_manager
        assert (
            b"http://stsci.edu/schemas/asdf/core/asdf-1.1.0"
            in config.resource_manager["http://stsci.edu/schemas/asdf/core/asdf-1.1.0"]
        )
        assert "http://somewhere.org/schemas/foo-1.0.0" not in config.resource_manager

        # Add a mapping and confirm that the manager now contains it:
        new_mapping = {"http://somewhere.org/schemas/foo-1.0.0": b"foo"}
        config.add_resource_mapping(new_mapping)
        assert "http://stsci.edu/schemas/asdf/core/asdf-1.1.0" in config.resource_manager
        assert (
            b"http://stsci.edu/schemas/asdf/core/asdf-1.1.0"
            in config.resource_manager["http://stsci.edu/schemas/asdf/core/asdf-1.1.0"]
        )
        assert "http://somewhere.org/schemas/foo-1.0.0" in config.resource_manager
        assert config.resource_manager["http://somewhere.org/schemas/foo-1.0.0"] == b"foo"

        # Remove a mapping and confirm that the manager no longer contains it:
        config.remove_resource_mapping(new_mapping)
        assert "http://stsci.edu/schemas/asdf/core/asdf-1.1.0" in config.resource_manager
        assert (
            b"http://stsci.edu/schemas/asdf/core/asdf-1.1.0"
            in config.resource_manager["http://stsci.edu/schemas/asdf/core/asdf-1.1.0"]
        )
        assert "http://somewhere.org/schemas/foo-1.0.0" not in config.resource_manager

        # Reset and confirm that the manager no longer contains the custom mapping:
        config.add_resource_mapping(new_mapping)
        config.reset_resources()
        assert "http://stsci.edu/schemas/asdf/core/asdf-1.1.0" in config.resource_manager
        assert (
            b"http://stsci.edu/schemas/asdf/core/asdf-1.1.0"
            in config.resource_manager["http://stsci.edu/schemas/asdf/core/asdf-1.1.0"]
        )
        assert "http://somewhere.org/schemas/foo-1.0.0" not in config.resource_manager


def test_extensions():
    with asdf.config_context() as config:
        original_extensions = config.extensions

        class BarExtension:
            extension_uri = "asdf://somewhere.org/extensions/bar-1.0"
            types = []
            tag_mapping = []
            url_mapping = []

        uri_extension = BarExtension()

        # Add an extension:
        config.add_extension(uri_extension)
        assert len(config.extensions) == len(original_extensions) + 1
        assert any(e for e in config.extensions if e.delegate is uri_extension)

        # Adding an extension should be idempotent:
        config.add_extension(uri_extension)
        assert len(config.extensions) == len(original_extensions) + 1

        # Even when wrapped:
        config.add_extension(ExtensionProxy(uri_extension))
        assert len(config.extensions) == len(original_extensions) + 1

        # Remove an extension:
        config.remove_extension(uri_extension)
        assert len(config.extensions) == len(original_extensions)

        # Removing should work when wrapped:
        config.add_extension(uri_extension)
        config.remove_extension(ExtensionProxy(uri_extension))
        assert len(config.extensions) == len(original_extensions)

        # And also by URI:
        config.add_extension(uri_extension)
        config.remove_extension(uri_extension.extension_uri)
        assert len(config.extensions) == len(original_extensions)

        # And also by URI pattern:
        config.add_extension(uri_extension)
        config.remove_extension("asdf://somewhere.org/extensions/*")
        assert len(config.extensions) == len(original_extensions)

        # Remove by the name of the extension's package:
        config.add_extension(ExtensionProxy(uri_extension, package_name="foo"))
        config.remove_extension(package="foo")
        assert len(config.extensions) == len(original_extensions)

        # Can combine remove filters:
        config.add_extension(ExtensionProxy(uri_extension, package_name="foo"))
        config.add_extension(ExtensionProxy(uri_extension, package_name="bar"))
        config.remove_extension(uri_extension.extension_uri, package="foo")
        assert len(config.extensions) == len(original_extensions) + 1

        # ... but not omit both:
        with pytest.raises(ValueError, match=r"Must specify at least one of extension or package"):
            config.remove_extension()

        # Removing an extension should be idempotent:
        config.add_extension(uri_extension)
        config.remove_extension(uri_extension)
        config.remove_extension(uri_extension)
        assert len(config.extensions) == len(original_extensions)

        # Resetting should get rid of any additions:
        config.add_extension(uri_extension)
        config.reset_extensions()
        assert len(config.extensions) == len(original_extensions)


def test_config_repr():
    with asdf.config_context() as config:
        config.validate_on_read = True
        config.default_version = "1.5.0"
        config.io_block_size = 9999
        config.legacy_fill_schema_defaults = False
        config.array_inline_threshold = 14

        assert "validate_on_read: True" in repr(config)
        assert "default_version: 1.5.0" in repr(config)
        assert "io_block_size: 9999" in repr(config)
        assert "legacy_fill_schema_defaults: False" in repr(config)
        assert "array_inline_threshold: 14" in repr(config)


@pytest.mark.parametrize("value", [True, False])
def test_get_set_default_array_save_base(value):
    with asdf.config_context() as config:
        config.default_array_save_base = value
        assert config.default_array_save_base == value


@pytest.mark.parametrize("value", [1, None])
def test_invalid_set_default_array_save_base(value):
    with asdf.config_context() as config:
        with pytest.raises(ValueError, match="default_array_save_base must be a bool"):
            config.default_array_save_base = value



"""
This file manages a transient representation of the tree made up of
simple Python data types (lists, dicts, scalars) wrapped inside of
`Tagged` subclasses, which add a ``tag`` attribute to hold the
associated YAML tag.

Below "basic data types" refers to the basic built-in data types
defined in the core YAML specification.  "Custom data types" are
specialized tags that are added by ASDF or third-parties that are not
in the YAML specification.

When YAML is loaded from disk, we want to first validate it using JSON
schema, which only understands basic Python data types, not the
``Nodes`` that ``pyyaml`` uses as its intermediate representation.
However, basic Python data types do not preserve the tag information
from the YAML file that we need later to convert elements to custom
data types.  Therefore, the approach here is to wrap those basic types
inside of `Tagged` objects long enough to run through the asdf._jsonschema
validator, and then convert to custom data types and throwing away the
tag annotations in the process.

Upon writing, the custom data types are first converted to basic
Python data types wrapped in `Tagged` objects.  The tags assigned to
the ``Tagged`` objects are then used to write tags to the YAML file.

All of this is an implementation detail of the our custom YAML loader
and dumper (``yamlutil.AsdfLoader`` and ``yamlutil.AsdfDumper``) and
is not intended to be exposed to the end user.
"""

from collections import UserDict, UserList, UserString
from copy import copy, deepcopy

__all__ = ["Tagged", "TaggedDict", "TaggedList", "TaggedString", "get_tag", "tag_object"]


class Tagged:
    """
    Base class of classes that wrap a given object and store a tag
    with it.
    """

    _base_type = None

    @property
    def base(self):
        """Convert to base type"""

        return self._base_type(self)


class TaggedDict(Tagged, UserDict, dict):
    """
    A Python dict with a tag attached.
    """

    _base_type = dict

    flow_style = None
    property_order = None

    def __init__(self, data=None, tag=None):
        if data is None:
            data = {}
        self.data = data
        self._tag = tag

    def __eq__(self, other):
        return isinstance(other, TaggedDict) and self.data == other.data and self._tag == other._tag

    def __deepcopy__(self, memo):
        data_copy = deepcopy(self.data, memo)
        return TaggedDict(data_copy, self._tag)

    def __copy__(self):
        data_copy = copy(self.data)
        return TaggedDict(data_copy, self._tag)


class TaggedList(Tagged, UserList, list):
    """
    A Python list with a tag attached.
    """

    _base_type = list

    flow_style = None

    def __init__(self, data=None, tag=None):
        if data is None:
            data = []
        self.data = data
        self._tag = tag

    def __eq__(self, other):
        return isinstance(other, TaggedList) and self.data == other.data and self._tag == other._tag

    def __deepcopy__(self, memo):
        data_copy = deepcopy(self.data, memo)
        return TaggedList(data_copy, self._tag)

    def __copy__(self):
        data_copy = copy(self.data)
        return TaggedList(data_copy, self._tag)


class TaggedString(Tagged, UserString, str):
    """
    A Python string with a tag attached.
    """

    _base_type = str

    style = None

    def __eq__(self, other):
        return isinstance(other, TaggedString) and str.__eq__(self, other) and self._tag == other._tag


def tag_object(tag, instance, ctx=None):
    """
    Tag an object by wrapping it in a ``Tagged`` instance.
    """
    if isinstance(instance, Tagged):
        instance._tag = tag
    elif isinstance(instance, dict):
        instance = TaggedDict(instance, tag)
    elif isinstance(instance, list):
        instance = TaggedList(instance, tag)
    elif isinstance(instance, str):
        instance = TaggedString(instance)
        instance._tag = tag
    else:
        from . import AsdfFile, yamlutil

        if ctx is None:
            ctx = AsdfFile()
        try:
            instance = yamlutil.custom_tree_to_tagged_tree(instance, ctx)
        except TypeError as err:
            msg = f"Don't know how to tag a {type(instance)}"
            raise TypeError(msg) from err
        instance._tag = tag
    return instance


def get_tag(instance):
    """
    Get the tag associated with the instance, if there is one.
    """
    if not isinstance(instance, Tagged):
        return None
    return getattr(instance, "_tag", None)



import numpy as np

import asdf


def test_control_array_storage_in_to_yaml_tree_methods(tmp_path):
    """
    controlling array storage in to_yaml_tree methods

    https://github.com/asdf-format/asdf/issues/1013
    """

    class FooType:
        def __init__(self, data):
            self.data = data

    class FooConverter:
        tags = ["asdf://somewhere.org/tag/foo-1.0.0"]
        types = [FooType]

        def to_yaml_tree(self, obj, tag, ctx):
            if obj.data.ndim < 2:
                ctx._blocks._set_array_storage(obj.data, "inline")
            return {"data": obj.data}

        def from_yaml_tree(self, obj, tag, ctx):
            return FooType(obj["data"])

    class FooExtension:
        converters = [FooConverter()]
        tags = ["asdf://somewhere.org/tag/foo-1.0.0"]
        extension_uri = "asdf://somewhere.org/extensions/foo-1.0.0"

    with asdf.config_context() as cfg:
        cfg.add_extension(FooExtension())

        fn = tmp_path / "test.asdf"

        for shape in [3, (3, 3)]:
            arr = np.zeros(shape)
            n_blocks = 0 if arr.ndim == 1 else 1
            af = asdf.AsdfFile({"foo": FooType(arr)})
            assert af.get_array_storage(arr) == "internal"
            af.write_to(fn)
            # make sure write_to doesn't change the settings outside of the
            # writing context
            assert af.get_array_storage(arr) == "internal"

            with asdf.open(fn) as af:
                np.testing.assert_array_equal(af["foo"].data, arr)
                assert len(af._blocks.blocks) == n_blocks



from .ndarray import numpy_dtype_to_asdf_datatype


class Stream:
    """
    Used to put a streamed array into the tree.

    Examples
    --------
    Save a double-precision array with 1024 columns, one row at a
    time::

         >>> from asdf import AsdfFile, Stream
         >>> import numpy as np
         >>> ff = AsdfFile()
         >>> ff.tree['streamed'] = Stream([1024], np.float64)
         >>> with open('test.asdf', 'wb') as fd:
         ...     ff.write_to(fd)
         ...     for i in range(200):
         ...         nbytes = fd.write(
         ...                      np.array([i] * 1024, np.float64).tobytes())
    """

    def __init__(self, shape, dtype, strides=None):
        self._shape = shape
        self._datatype, self._byteorder = numpy_dtype_to_asdf_datatype(dtype)
        self._strides = strides
        self._array = None

    def _make_array(self):
        self._array = None

    def __repr__(self):
        return f"Stream({self._shape}, {self._datatype}, strides={self._strides})"

    def __str__(self):
        return str(self.__repr__())



import sys
import warnings

from .exceptions import AsdfWarning
from .extension import ExtensionProxy
from .resource import ResourceMappingProxy

# The standard library importlib.metadata returns duplicate entrypoints
# for all python versions up to and including 3.11
# https://github.com/python/importlib_metadata/issues/410#issuecomment-1304258228
# see PR https://github.com/asdf-format/asdf/pull/1260
# see issue https://github.com/asdf-format/asdf/issues/1254
if sys.version_info >= (3, 12):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points


RESOURCE_MAPPINGS_GROUP = "asdf.resource_mappings"
EXTENSIONS_GROUP = "asdf.extensions"


def get_resource_mappings():
    return _list_entry_points(RESOURCE_MAPPINGS_GROUP, ResourceMappingProxy)


def get_extensions():
    extensions = _list_entry_points(EXTENSIONS_GROUP, ExtensionProxy)
    return extensions


def _list_entry_points(group, proxy_class):
    results = []

    points = entry_points(group=group)

    # The order of plugins may be significant, since in the case of
    # duplicate functionality the first plugin in the list takes
    # precedence.  It's not clear if entry points are ordered
    # in a consistent way across systems so we explicitly sort
    # by package name.  Plugins from this package are placed
    # at the end so that other packages can override them.
    asdf_entry_points = [e for e in points if e.dist.name == "asdf"]
    other_entry_points = sorted((e for e in points if e.dist.name != "asdf"), key=lambda e: e.dist.name)

    for entry_point in other_entry_points + asdf_entry_points:
        package_name = entry_point.dist.name
        package_version = entry_point.dist.version

        def _handle_error(e):
            warnings.warn(
                f"{group} plugin from package {package_name}=={package_version} failed to load:\n\n"
                f"{e.__class__.__name__}: {e}",
                AsdfWarning,
            )

        # Catch errors loading entry points and warn instead of raising
        try:
            with warnings.catch_warnings():
                elements = entry_point.load()()

        except Exception as e:
            _handle_error(e)
            continue

        # Process the elements returned by the entry point
        if not isinstance(elements, list):
            elements = [elements]

        for element in elements:
            # Catch errors instantiating the proxy class and warn instead of raising
            try:
                results.append(proxy_class(element, package_name=package_name, package_version=package_version))
            except Exception as e:
                _handle_error(e)

    return results



"""
Low-level functions for reading and writing ASDF blocks
and other block related file contents (like the block index).
"""

import hashlib
import io
import os
import struct
import weakref

import yaml

from asdf import _compression as mcompression
from asdf import constants, generic_io, util
from asdf.versioning import _yaml_base_loader as BaseLoader

from .exceptions import BlockIndexError

BLOCK_HEADER = util._BinaryStruct(
    [
        ("flags", "I"),
        ("compression", "4s"),
        ("allocated_size", "Q"),
        ("used_size", "Q"),
        ("data_size", "Q"),
        ("checksum", "16s"),
    ],
)


def calculate_block_checksum(data):
    # The following line is safe because we're only using
    # the MD5 as a checksum.
    m = hashlib.new("md5", usedforsecurity=False)
    m.update(data)
    return m.digest()


def validate_block_header(header):
    """
    Check that they key value pairs in header contain consistent
    information about the ASDF block ``compression``, ``flags``,
    ``used_size`` and ``data_size`` (otherwise raise an exception).

    Parameters
    ----------
    header : dict
        ASDF block header information.

    Raises
    ------
    ValueError
        If the key value pairs in header contain inconsistent information
    """
    compression = mcompression.validate(header["compression"])
    if header["flags"] & constants.BLOCK_FLAG_STREAMED:
        if compression is not None:
            msg = "Compression set on a streamed block."
            raise ValueError(msg)
    else:
        if compression is None and header["used_size"] != header["data_size"]:
            msg = "used_size and data_size must be equal when no compression is used."
            raise ValueError(msg)
    return header


def read_block_header(fd, offset=None):
    """
    Read an ASDF block header

    Parameters
    ----------
    fd : file or generic_io.GenericIO
        File to read.

    offset : int, optional
        Offset within the file where the start of the ASDF block
        header is located. If provided, the file will be seeked prior
        to reading.

    Returns
    -------
    header : dict
        Dictionary containing the read ASDF header as parsed by the
        `BLOCK_HEADER` `asdf.util._BinaryStruct`.

    Raises
    ------
    ValueError
        If the read header is inconsistent (see `validate_block_header`).
    """
    if offset is not None:
        fd.seek(offset)

    # read the header size
    buff = fd.read(2)
    header_size = struct.unpack(b">H", buff)[0]
    if header_size < BLOCK_HEADER.size:
        msg = f"Header size must be >= {BLOCK_HEADER.size}"
        raise ValueError(msg)

    header = BLOCK_HEADER.unpack(fd.read(header_size))
    return validate_block_header(header)


def read_block_data(fd, header, validate_checksum, offset=None, memmap=False):
    """
    Read (or memory map) data for an ASDF block.

    Parameters
    ----------
    fd : file or generic_io.GenericIO
        File to read.

    header : dict
        ASDF block header dictionary (as read from `read_block_header`).

    validate_checksum: bool
        If `True`, raise an exception if the checksum in the block header
        doesn't match the checksum computed from the block body.
        Has no effect if the header doesn't contain a checksum.

    offset : int, optional
        Offset within the file where the start of the ASDF block data
        is located. If provided, the file will be seeked prior to reading.

    memmap : bool, optional, default False
        Memory map the block data using `generic_io.GenericIO.memmap_array`.
        A compressed block will never be memmapped and if the file ``fd``
        does not support memmapping the data will not be memmapped (and
        no error will be raised).

    Returns
    -------
    data : ndarray or memmap
        A one-dimensional ndarray of dtype uint8

    Raises
    ------
    ValueError
        If `validate_checksum` is set and the header checksum doesn't match
        the checksum computed from the block data.
    """

    if fd.seekable():
        if offset is not None:
            fd.seek(offset)
        else:
            offset = fd.tell()

    streamed = header["flags"] & constants.BLOCK_FLAG_STREAMED

    if streamed:
        used_size = -1
    else:
        used_size = header["used_size"]

    # if no compression, just read data
    compression = mcompression.validate(header["compression"])
    has_checksum = any(b != 0 for b in header["checksum"])

    if compression:
        if validate_checksum and has_checksum:
            cmp_data = fd.read(header["used_size"])
            # Fast-forward first so if we raise an exception the file pointer is still correct
            fd.fast_forward(header["allocated_size"] - header["used_size"])

            checksum = calculate_block_checksum(cmp_data)
            if header["checksum"] != checksum:
                msg = f"Block at {offset} does not match given checksum"
                raise ValueError(msg)

            data = mcompression.decompress(
                generic_io.get_file(io.BytesIO(cmp_data)), used_size, header["data_size"], compression
            )
        else:
            # compressed data will not be memmapped
            data = mcompression.decompress(fd, used_size, header["data_size"], compression)
            fd.fast_forward(header["allocated_size"] - header["used_size"])
    else:
        if memmap and fd.can_memmap():
            data = fd.memmap_array(offset, used_size)
            ff_bytes = header["allocated_size"]
        else:
            data = fd.read_into_array(used_size)
            ff_bytes = header["allocated_size"] - header["used_size"]
        if streamed and fd.seekable():
            fd.seek(0, os.SEEK_END)
        else:
            fd.fast_forward(ff_bytes)

        if validate_checksum and has_checksum:
            checksum = calculate_block_checksum(data)
            if header["checksum"] != checksum:
                msg = f"Block at {offset} does not match given checksum"
                raise ValueError(msg)
    return data


def read_block(fd, validate_checksum, offset=None, memmap=False, lazy_load=False):
    """
    Read a block (header and data) from an ASDF file.

    Parameters
    ----------
    fd : file or generic_io.GenericIO
        File to read.

    validate_checksum: bool
        If True, raise an exception if the checksum in the block header
        doesn't match the checksum computed from the block body.
        Has no effect if the header doesn't contain a checksum.

    offset : int, optional
        Offset within the file where the start of the ASDF block header
        is located. If provided, the file will be seeked prior to reading.
        Note this is the start of the block header not the start of the
        block magic.

    memmap : bool, optional, default False
        Memory map the block data see `read_block_data` for more
        details.

    lazy_load : bool, optional, default False
        Return a callable that when called will read the block data. This
        option is ignored for a non-seekable file.

    Returns
    -------
    offset : int
        The offset within the file where the block was read (equal to offset
        argument if it was provided).

    header : dict
        ASDF block header as read with `read_block_header`.

    data_offset : int
        The offset within the file where the block data begins.

    data : ndarray, memmap or callable
        ASDF block data (one-dimensional ndarray of dtype uint8). If lazy_load
        (and the file is seekable) data will be a callable that when executed
        will seek the file and read the block data.
    """
    # expects the fd or offset is past the block magic
    if offset is None and fd.seekable():
        offset = fd.tell()
    header = read_block_header(fd, offset)
    if fd.seekable():
        data_offset = fd.tell()
    else:
        data_offset = None
    if lazy_load and fd.seekable():
        # setup a callback to later load the data
        fd_ref = weakref.ref(fd)

        def callback():
            fd = fd_ref()
            if fd is None or fd.is_closed():
                msg = "ASDF file has already been closed. Can not get the data."
                raise OSError(msg)
            position = fd.tell()
            data = read_block_data(fd, header, validate_checksum, offset=data_offset, memmap=memmap)
            fd.seek(position)
            return data

        data = callback
        if header["flags"] & constants.BLOCK_FLAG_STREAMED:
            fd.seek(0, os.SEEK_END)
        else:
            fd.fast_forward(header["allocated_size"])
    else:
        data = read_block_data(fd, header, validate_checksum, offset=None, memmap=memmap)
    return offset, header, data_offset, data


def generate_write_header(
    data, stream=False, compression_kwargs=None, padding=False, fs_block_size=1, write_checksum=True, **header_kwargs
):
    """
    Generate a dict representation of a ASDF block header that can be
    used for writing a block.

    Note that if a compression key is provided in ``header_kwargs`` this
    function will compress ``data`` to determine the used_size (the
    compressed data will be returned via the ``buff`` result to avoid
    needing to re-compress the data before writing).

    Parameters
    ----------

    data : ndarray
        A one-dimensional ndarray of dtype uint8.

    stream : bool, optional, default False
        If True, generate a header for a streamed block.

    compression_kwargs : dict, optional
        If provided, these will be passed on to `asdf.compression.compress`
        if the data is compressed (see header_kwargs).

    padding : bool or float, optional, default False
        If the block should contain additional padding bytes. See the
        `asdf.util.calculate_padding` argument ``pad_blocks`` for more
        details.

    fs_block_size : int, optional, default 1
        The filesystem block size. See the `asdf.util.calculate_padding`
        ``block_size`` argument for more details.

    write_checksum: bool, optional
        Compute and write the checksum of the block data.
        If disabled then the checksum field is set to 0.

    **header_kwargs : dict, optional
        Block header settings that will be read, updated, and used
        to generate the binary block header representation by packing
        with `BLOCK_HEADER`.

    Returns
    -------

    header : dict
        Dictionary representation of an ASDF block header.

    buff : bytes or None
        If this block is compressed buff will contained the compressed
        representation of data or None if the data is uncompressed.

    padding_bytes: int
        The number of padding bytes that must be written after
        the block data.
    """
    if data.ndim != 1 or data.dtype != "uint8":
        msg = "Data must be of ndim==1 and dtype==uint8"
        raise ValueError(msg)
    if stream:
        header_kwargs["flags"] = header_kwargs.get("flags", 0) | constants.BLOCK_FLAG_STREAMED
        header_kwargs["data_size"] = 0
    else:
        header_kwargs["flags"] = 0
        header_kwargs["data_size"] = data.nbytes

    header_kwargs["compression"] = mcompression.to_compression_header(header_kwargs.get("compression", None))

    if header_kwargs["compression"] == b"\0\0\0\0":
        used_size = header_kwargs["data_size"]
        buff = None
    else:
        buff = io.BytesIO()
        mcompression.compress(buff, data, header_kwargs["compression"], config=compression_kwargs)
        used_size = buff.tell()
    if stream:
        header_kwargs["used_size"] = 0
        header_kwargs["allocated_size"] = 0
    else:
        header_kwargs["used_size"] = used_size
        padding = util.calculate_padding(used_size, padding, fs_block_size)
        header_kwargs["allocated_size"] = header_kwargs.get("allocated_size", used_size + padding)

    if stream or not write_checksum:
        header_kwargs["checksum"] = b"\0" * 16
    elif buff is not None:
        header_kwargs["checksum"] = calculate_block_checksum(buff.getbuffer())
    else:
        header_kwargs["checksum"] = calculate_block_checksum(data)

    if header_kwargs["allocated_size"] < header_kwargs["used_size"]:
        msg = (
            f"Block used size {header_kwargs['used_size']} larger than "
            f"allocated size {header_kwargs['allocated_size']}",
        )
        raise RuntimeError(msg)
    padding_bytes = header_kwargs["allocated_size"] - header_kwargs["used_size"]
    return header_kwargs, buff, padding_bytes


def write_block(
    fd, data, offset=None, stream=False, compression_kwargs=None, padding=False, write_checksum=True, **header_kwargs
):
    """
    Write an ASDF block.

    Parameters
    ----------
    fd : file or generic_io.GenericIO
        File to write to.

    offset : int, optional
        If provided, seek to this offset before writing.

    stream : bool, optional, default False
        If True, write this as a streamed block.

    compression_kwargs : dict, optional
        If block is compressed, use these additional arguments during
        compression. See `generate_write_header`.

    padding : bool, optional, default False
        Optionally pad the block data. See `generate_write_header`.

    write_checksum: bool, optional
        Compute and write the checksum of the block data.
        If disabled then the checksum field is set to 0.

    **header_kwargs : dict
        Block header settings. See `generate_write_header`.

    Returns
    -------

    header : dict
        The ASDF block header as unpacked from the `BLOCK_HEADER` used
        for writing.
    """
    header_dict, buff, padding_bytes = generate_write_header(
        data, stream, compression_kwargs, padding, fd.block_size, write_checksum, **header_kwargs
    )
    header_bytes = BLOCK_HEADER.pack(**header_dict)

    if offset is not None:
        if fd.seekable():
            fd.seek(offset)
        else:
            msg = "write_block received offset for non-seekable file"
            raise ValueError(msg)
    fd.write(struct.pack(b">H", len(header_bytes)))
    fd.write(header_bytes)
    if buff is None:  # data is uncompressed
        fd.write_array(data)
    else:
        fd.write(buff.getbuffer())
    fd.fast_forward(padding_bytes)
    return header_dict


def _candidate_offsets(min_offset, max_offset, block_size):
    offset = (max_offset // block_size) * block_size
    if offset == max_offset:
        offset -= block_size
    while offset > min_offset:
        yield offset
        offset -= block_size
    if offset <= min_offset:
        yield min_offset


def find_block_index(fd, min_offset=None, max_offset=None):
    """
    Find the location of an ASDF block index within a seekable file.

    Searching will begin at the end of the file (or max_offset
    if it is provided).

    Parameters
    ----------

    fd : file or generic_io.GenericIO
        A seekable file that will be searched to try and find
        the start of an ASDF block index within the file.

    min_offset : int, optional
        The minimum search offset. A block index will not be
        found before this point.

    max_offset : int, optional
        The maximum search offset. A block index will not be
        found after this point.

    Returns
    -------

    offset : int or None
        Index of start of ASDF block index. This is the location of the
        ASDF block index header.

    """
    if min_offset is None:
        min_offset = fd.tell()
    if max_offset is None:
        fd.seek(0, os.SEEK_END)
        max_offset = fd.tell()
    block_size = fd.block_size
    block_index_offset = None
    buff = b""
    pattern = constants.INDEX_HEADER
    for offset in _candidate_offsets(min_offset, max_offset, block_size):
        fd.seek(offset)
        buff = fd.read(block_size) + buff
        index = buff.find(pattern)
        if index != -1:
            block_index_offset = offset + index
            if block_index_offset >= max_offset:
                return None
            break
        buff = buff[: len(pattern)]
    return block_index_offset


def read_block_index(fd, offset=None):
    """
    Read an ASDF block index from a file.

    Parameters
    ----------

    fd : file or generic_io.GenericIO
        File to read the block index from.

    offset : int, optional
        Offset within the file where the block index starts
        (the start of the ASDF block index header). If not provided
        reading will start at the current position of the file
        pointer. See `find_block_index` to locate the block
        index prior to calling this function.

    Returns
    -------

    block_index : list of ints
        A list of ASDF block offsets read and parsed from the
        block index.

    Raises
    ------
    BlockIndexError
        The data read from the file did not contain a valid
        block index.
    """
    if offset is not None:
        fd.seek(offset)
    buff = fd.read(len(constants.INDEX_HEADER))
    if buff != constants.INDEX_HEADER:
        msg = "Failed to read block index header at offset {offset}"
        raise BlockIndexError(msg)
    try:
        # the noqa is needed here since the linter doesn't know
        # that BaseLoader here is either SafeLoader or CSafeLoader
        # both of which do not violate S506.
        block_index = yaml.load(fd.read(-1), BaseLoader)  # noqa: S506
    except yaml.error.YAMLError:
        raise BlockIndexError("Failed to parse block index as yaml")
    if (
        not isinstance(block_index, list)
        or any(not isinstance(v, int) for v in block_index)
        or block_index != sorted(block_index)
    ):
        raise BlockIndexError("Invalid block index")
    return block_index


def write_block_index(fd, offsets, offset=None, yaml_version=None):
    """
    Write a list of ASDF block offsets to a file in the form
    of an ASDF block index.

    Parameters
    ----------
    fd : file or generic_io.GenericIO
        File to write to.

    offsets : list of ints
        List of byte offsets (from the start of the file) where
        ASDF blocks are located.

    offset : int, optional
        If provided, seek to this offset before writing.

    yaml_version : tuple, optional, default (1, 1)
        YAML version to use when writing the block index. This
        will be passed to ``yaml.dump`` as the version argument.
    """
    if yaml_version is None:
        yaml_version = (1, 1)
    if offset is not None:
        fd.seek(offset)
    fd.write(constants.INDEX_HEADER)
    fd.write(b"\n")
    yaml.dump(
        offsets,
        stream=fd,
        Dumper=yaml.SafeDumper,
        explicit_start=True,
        explicit_end=True,
        allow_unicode=True,
        encoding="utf-8",
        version=yaml_version,
    )



"""
This module deals with things that change between different versions
of the ASDF spec.
"""

from functools import total_ordering

import yaml
from semantic_version import SimpleSpec, Version

try:
    from asdf_standard._versioning import get_supported_core_schema_versions
except ImportError:

    def get_supported_core_schema_versions():
        return ("1.0.0", "1.1.0", "1.2.0", "1.3.0", "1.4.0", "1.5.0", "1.6.0")


_yaml_base_loader = yaml.CSafeLoader if getattr(yaml, "__with_libyaml__", None) else yaml.SafeLoader


__all__ = ["AsdfVersion", "AsdfVersionMixin", "join_tag_version", "split_tag_version"]


def split_tag_version(tag):
    """
    Split a tag into its base and version.
    """
    name, version = tag.rsplit("-", 1)
    version = AsdfVersion(version)
    return name, version


def join_tag_version(name, version):
    """
    Join the root and version of a tag back together.
    """
    return f"{name}-{version}"


@total_ordering
class AsdfVersionMixin:
    """This mix-in is required in order to impose the total ordering that we
    want for ``AsdfVersion``, rather than accepting the total ordering that is
    already provided by ``Version`` from ``semantic_version``. Defining these
    comparisons directly in ``AsdfVersion`` and applying ``total_ordering``
    there will not work since ``total_ordering`` only defines comparison
    operations if they do not exist already and the base class ``Version``
    already defines these operations.
    """

    def __eq__(self, other):
        # Seems like a bit of a hack...
        if isinstance(other, SimpleSpec):
            return other == self
        if isinstance(other, (str, tuple, list)):
            other = AsdfVersion(other)
        return Version.__eq__(self, other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, (str, tuple, list)):
            other = AsdfVersion(other)
        return Version.__lt__(self, other)

    def __hash__(self):
        # To be honest, I'm not sure why I had to make this explicit
        return Version.__hash__(self)


class AsdfVersion(AsdfVersionMixin, Version):
    """This class adds features to the existing ``Version`` class from the
    ``semantic_version`` module. Namely, it allows ``Version`` objects to be
    constructed from tuples and lists as well as strings, and it allows
    ``Version`` objects to be compared with tuples, lists, and strings, instead
    of just other ``Version`` objects.

    If any of these features are added to the ``Version`` class itself (as
    requested in https://github.com/rbarrois/python-semanticversion/issues/52),
    then this class will become obsolete.
    """

    def __init__(self, version):
        # This is a dirty hack and you know it
        if isinstance(version, AsdfVersion):
            version = str(version)
        if isinstance(version, (tuple, list)):
            version = ".".join([str(x) for x in version])
        super().__init__(version)


supported_versions = tuple(AsdfVersion(version) for version in get_supported_core_schema_versions())

default_version = supported_versions[-1]


# This is the ASDF core schemas version at which the format of the history
# field changed to include extension metadata.
NEW_HISTORY_FORMAT_MIN_VERSION = AsdfVersion("1.2.0")


# This is the ASDF core schemas version at which we begin restricting
# mapping keys to string, integer, and boolean only.
RESTRICTED_KEYS_MIN_VERSION = AsdfVersion("1.6.0")


# This library never removed defaults for ASDF core schemas versions
# later than 1.5.0, so filling them isn't necessary.
FILL_DEFAULTS_MAX_VERSION = AsdfVersion("1.5.0")

# ASDF currently only defined a single file format version
_FILE_FORMAT_VERSION = AsdfVersion("1.0.0")

# ASDF currently only supports a single yaml version
# use a tuple as that is what yaml expects
_YAML_VERSION = (1, 1)



import importlib.resources

import numpy as np
import pytest

from asdf import config, schema

from .httpserver import HTTPServer


@pytest.fixture()
def small_tree():
    x = np.arange(0, 10, dtype=float)
    return {
        "science_data": x,
        "subset": x[3:-3],
        "skipping": x[::2],
        "not_shared": np.arange(10, 0, -1, dtype=np.uint8),
    }


@pytest.fixture()
def large_tree():
    # These are designed to be big enough so they don't fit in a
    # single block, but not so big that RAM/disk space for the tests
    # is enormous.
    x = np.zeros((256, 256))
    y = np.ones((16, 16, 16))
    return {
        "science_data": x,
        "more": y,
    }


@pytest.fixture
def recursive_tree(small_tree):
    a = small_tree.copy()
    a["a"] = a
    return a


@pytest.fixture(
    params=[
        "small_tree",
        "large_tree",
        "recursive_tree",
    ]
)
def tree(request):
    """
    Metafixture for all tree fixtures.
    """
    return request.getfixturevalue(request.param)


@pytest.fixture(autouse=True)
def _restore_default_config():
    yield
    config._global_config = config.AsdfConfig()
    config._local = config._ConfigLocal()


@pytest.fixture(autouse=True)
def _clear_schema_cache():
    """
    Fixture that clears schema caches to prevent issues
    when tests use same URI for different schema content.
    """
    yield
    schema._load_schema.cache_clear()
    schema._load_schema_cached.cache_clear()


@pytest.fixture()
def httpserver(request):
    """
    The returned ``httpserver`` provides a threaded HTTP server
    instance.  It serves content from a temporary directory (available
    as the attribute tmpdir) at randomly assigned URL (available as
    the attribute url).

    * ``tmpdir`` - path to the tmpdir that it's serving from (str)
    * ``url`` - the base url for the server
    """
    server = HTTPServer()
    yield server
    server.finalize()


@pytest.fixture()
def test_data_path():
    return importlib.resources.files("asdf") / "_tests" / "data"


@pytest.fixture(params=[True, False], ids=["lazy", "not-lazy"])
def with_lazy_tree(request):
    with config.config_context() as cfg:
        cfg.lazy_tree = request.param
        yield cfg.lazy_tree



import collections

from .constant import Constant
from .external_reference import ExternalArrayReference
from .integer import IntegerType
from .ndarray import NDArrayType
from .stream import Stream

__all__ = [
    "AsdfObject",
    "Constant",
    "ExtensionMetadata",
    "ExternalArrayReference",
    "HistoryEntry",
    "IntegerType",
    "NDArrayType",
    "Software",
    "Stream",
    "SubclassMetadata",
]


# AsdfObject inherits both collections.UserDict and dict to allow it
# to pass an isinstance(..., dict) check and to allow it to be "lazy"
# loaded when "lazy_tree=True".
class AsdfObject(collections.UserDict, dict):
    pass


class Software(dict):
    pass


class HistoryEntry(dict):
    pass


class ExtensionMetadata(dict):
    @property
    def extension_uri(self):
        return self.get("extension_uri")

    @property
    def extension_class(self):
        return self["extension_class"]

    @property
    def software(self):
        return self.get("software")


class SubclassMetadata(dict):
    """
    The tagged object supported by this class is part of
    an experimental feature that has since been dropped
    from this library.  This class remains so that ASDF
    files that used that feature will still deserialize
    without warnings.
    """



"""
Support for plugins that provide access to resources such
as schemas.
"""

import pkgutil
from collections.abc import Mapping

from asdf_standard import DirectoryResourceMapping as _DirectoryResourceMapping

from .util import get_class_name

__all__ = [
    "DirectoryResourceMapping",
    "JsonschemaResourceMapping",
    "ResourceManager",
    "ResourceMappingProxy",
]


class DirectoryResourceMapping(_DirectoryResourceMapping):
    """
    A resource mapping that reads resource content from a directory or directory tree.

    See :class:`~asdf_standard.resource.DirectoryResourceMapping` for details.
    """


class ResourceMappingProxy(Mapping):
    """
    Wrapper around a resource mapping that carries
    additional information on the package that provided
    the mapping.
    """

    @classmethod
    def maybe_wrap(cls, delegate):
        if isinstance(delegate, ResourceMappingProxy):
            return delegate

        return ResourceMappingProxy(delegate)

    def __init__(self, delegate, package_name=None, package_version=None):
        if not isinstance(delegate, Mapping):
            msg = "Resource mapping must implement the Mapping interface"
            raise TypeError(msg)

        self._delegate = delegate
        self._package_name = package_name
        self._package_version = package_version
        self._class_name = get_class_name(delegate)

    def __getitem__(self, uri):
        return self._delegate.__getitem__(uri)

    def __len__(self):
        return self._delegate.__len__()

    def __iter__(self):
        return self._delegate.__iter__()

    @property
    def delegate(self):
        """
        Get the wrapped mapping instance.

        Returns
        -------
        collections.abc.Mapping
        """
        return self._delegate

    @property
    def package_name(self):
        """
        Get the name of the Python package that provided this mapping.

        Returns
        -------
        str or None
            `None` if the mapping was added at runtime.
        """
        return self._package_name

    @property
    def package_version(self):
        """
        Get the version of the Python package that provided the mapping.

        Returns
        -------
        str or None
            `None` if the mapping was added at runtime.
        """
        return self._package_version

    @property
    def class_name(self):
        """ "
        Get the fully qualified class name of the mapping.

        Returns
        -------
        str
        """
        return self._class_name

    def __eq__(self, other):
        if isinstance(other, ResourceMappingProxy):
            return other.delegate is self.delegate

        return False

    def __hash__(self):
        return hash(id(self.delegate))

    def __repr__(self):
        if self.package_name is not None:
            package_description = f"{self.package_name}=={self.package_version}"
        else:
            package_description = "(none)"

        return f"<ResourceMappingProxy class: {self.class_name} package: {package_description} len: {len(self)}>"


class ResourceManager(Mapping):
    """
    Wraps multiple resource mappings into a single interface
    with some friendlier error handling.

    Parameters
    ----------
    resource_mappings : iterable of collections.abc.Mapping
        Underlying resource mappings.  In the case of a duplicate URI,
        the first mapping takes precedence.
    """

    def __init__(self, resource_mappings):
        self._resource_mappings = resource_mappings

        self._mappings_by_uri = {}
        for mapping in resource_mappings:
            for uri in mapping:
                if uri not in self._mappings_by_uri:
                    self._mappings_by_uri[uri] = mapping

    def __getitem__(self, uri):
        if uri not in self._mappings_by_uri:
            msg = f"Resource unavailable for URI: {uri}"
            raise KeyError(msg)

        content = self._mappings_by_uri[uri][uri]
        if isinstance(content, str):
            content = content.encode("utf-8")

        return content

    def __len__(self):
        return len(self._mappings_by_uri)

    def __iter__(self):
        yield from self._mappings_by_uri

    def __contains__(self, uri):
        # Implement __contains__ only for efficiency.
        return uri in self._mappings_by_uri

    def __repr__(self):
        return f"<ResourceManager len: {self.__len__()}>"


_JSONSCHEMA_URI_TO_FILENAME = {
    "http://json-schema.org/draft-04/schema": "draft4.json",
}


class JsonschemaResourceMapping(Mapping):
    """
    Resource mapping that fetches metaschemas from
    the jsonschema package.
    """

    def __getitem__(self, uri):
        filename = _JSONSCHEMA_URI_TO_FILENAME[uri]
        return pkgutil.get_data("asdf._jsonschema", f"schemas/{filename}")

    def __len__(self):
        return len(_JSONSCHEMA_URI_TO_FILENAME)

    def __iter__(self):
        yield from _JSONSCHEMA_URI_TO_FILENAME

    def __repr__(self):
        return "JsonschemaResourceMapping()"



import gc
from unittest.mock import patch

import pytest

from asdf._block.key import Key
from asdf._block.store import Store


# a blank class for testing
class Foo:
    pass


def test_store_by_obj():
    f = Foo()
    v = 42
    s = Store()
    s.assign_object(f, v)
    assert s.lookup_by_object(f) == v


def test_get_missing_by_obj():
    f = Foo()
    s = Store()
    assert s.lookup_by_object(f) is None


def test_store_by_key():
    f = Foo()
    v = 42
    s = Store()
    k = Key(f)
    s.assign_object(k, v)
    assert s.lookup_by_object(k) == v


def test_get_by_key():
    f = Foo()
    v = 42
    s = Store()
    k = Key(f)
    s.assign_object(k, v)
    assert s.lookup_by_object(f) == v


def test_get_missing_key():
    f = Foo()
    s = Store()
    k = Key(f)
    assert s.lookup_by_object(k) is None


def test_get_missing_key_same_obj():
    f = Foo()
    v = 42
    s = Store()
    k = Key(f)
    s.assign_object(k, v)
    k2 = Key(f)
    assert s.lookup_by_object(k2) is None


def test_get_existing_default():
    f = Foo()
    v = 42
    s = Store()
    s.assign_object(f, v)
    assert s.lookup_by_object(f, 26) == v


def test_get_missing_default():
    f = Foo()
    v = 42
    s = Store()
    assert s.lookup_by_object(f, v) == v


def test_set_same_object():
    f = Foo()
    v = 42
    s = Store()
    s.assign_object(f, 26)
    s.assign_object(f, v)
    assert s.lookup_by_object(f) == v


def test_invalid_key_assign_object():
    s = Store()
    k = Key()
    with pytest.raises(ValueError, match="Invalid key used for assign_object"):
        s.assign_object(k, 42)


def test_set_same_key():
    f = Foo()
    s = Store()
    k = Key(f)
    v = 42
    s.assign_object(k, 26)
    s.assign_object(k, v)
    assert s.lookup_by_object(k) == v


def test_get_memory_reused():
    f = Foo()
    s = Store()
    v = 42
    s.assign_object(f, v)
    fid = id(f)
    del f
    gc.collect(2)
    f2 = Foo()

    def mock_id(obj):
        if obj is f2:
            return fid
        return id(obj)

    with patch("asdf._block.store.id", mock_id):
        assert s.lookup_by_object(f2) is None


def test_set_memory_reused():
    f = Foo()
    s = Store()
    v = 42
    s.assign_object(f, v)
    fid = id(f)
    del f
    gc.collect(2)
    f2 = Foo()

    def mock_id(obj):
        if obj is f2:
            return fid
        return id(obj)

    with patch("asdf._block.store.id", mock_id):
        nv = 26
        s.assign_object(f2, nv)
        assert s.lookup_by_object(f2) is nv


def test_cleanup():
    f = Foo()
    s = Store()
    k = Key(f)
    s.assign_object(s, 42)
    s.assign_object(k, 26)
    del f
    gc.collect(2)
    s._cleanup()
    assert s.lookup_by_object(k, None) is None


def test_keys_for_value():
    s = Store()
    data = {
        Foo(): 42,
        Foo(): 26,
        Foo(): 42,
        Foo(): 11,
    }
    data_by_value = {}
    for o, v in data.items():
        s.assign_object(o, v)
        data_by_value[v] = [*data_by_value.get(v, []), o]

    for v, objs in data_by_value.items():
        objs = set(objs)
        returned_objects = set()
        for k in s.keys_for_value(v):
            assert k._is_valid()
            obj = k._ref()
            returned_objects.add(obj)
        assert objs == returned_objects
        del returned_objects, objs
        gc.collect(2)



import argparse
from pathlib import Path

from common import assert_file_correct


def parse_args():
    parser = argparse.ArgumentParser(description="Confirm that an ASDF file generated by generate_file.py can be read")
    parser.add_argument("filename", help="the filename to test")
    return parser.parse_args()


def main():
    args = parse_args()

    path = Path(args.filename)

    assert_file_correct(path)


if __name__ == "__main__":
    main()



from copy import copy, deepcopy

import pytest

import asdf
from asdf.tagged import Tagged, TaggedDict, TaggedList, TaggedString, get_tag


def test_tagged_list_deepcopy():
    original = TaggedList([0, 1, 2, ["foo"]], "tag:nowhere.org:custom/foo-1.0.0")
    result = deepcopy(original)
    assert result == original
    assert result.data == original.data
    assert result._tag == original._tag
    original.append(4)
    assert len(result) == 4
    original[3].append("bar")
    assert len(result[3]) == 1


def test_tagged_list_copy():
    original = TaggedList([0, 1, 2, ["foo"]], "tag:nowhere.org:custom/foo-1.0.0")
    result = copy(original)
    assert result == original
    assert result.data == original.data
    assert result._tag == original._tag
    original.append(4)
    assert len(result) == 4
    original[3].append("bar")
    assert len(result[3]) == 2


def test_tagged_list_isinstance():
    value = TaggedList([0, 1, 2, ["foo"]], "tag:nowhere.org:custom/foo-1.0.0")
    assert isinstance(value, list)


def test_tagged_list_base():
    value = TaggedList([0, 1, 2, ["foo"]], "tag:nowhere.org:custom/foo-1.0.0")

    assert not (value == value.base)  # base is not a tagged list
    assert value.data == value.base  # but the data is

    assert isinstance(value.base, list)
    assert not isinstance(value.base, TaggedList)
    assert value.base.__class__ is list


def test_tagged_dict_deepcopy():
    original = TaggedDict({"a": 0, "b": 1, "c": 2, "nested": {"d": 3}}, "tag:nowhere.org:custom/foo-1.0.0")
    result = deepcopy(original)
    assert result == original
    assert result.data == original.data
    assert result._tag == original._tag
    original["e"] = 4
    assert len(result) == 4
    original["nested"]["f"] = 5
    assert len(result["nested"]) == 1


def test_tagged_dict_copy():
    original = TaggedDict({"a": 0, "b": 1, "c": 2, "nested": {"d": 3}}, "tag:nowhere.org:custom/foo-1.0.0")
    result = copy(original)
    assert result == original
    assert result.data == original.data
    assert result._tag == original._tag
    original["e"] = 4
    assert len(result) == 4
    original["nested"]["f"] = 5
    assert len(result["nested"]) == 2


def test_tagged_dict_isinstance():
    value = TaggedDict({"a": 0, "b": 1, "c": 2, "nested": {"d": 3}}, "tag:nowhere.org:custom/foo-1.0.0")
    assert isinstance(value, dict)


def test_tagged_dict_base():
    value = TaggedDict({"a": 0, "b": 1, "c": 2, "nested": {"d": 3}}, "tag:nowhere.org:custom/foo-1.0.0")

    assert not (value == value.base)  # base is not a tagged  dict
    assert value.data == value.base  # but the data is

    assert isinstance(value.base, dict)
    assert not isinstance(value.base, TaggedDict)
    assert value.base.__class__ is dict


def test_tagged_string_deepcopy():
    original = TaggedString("You're it!")
    original._tag = "tag:nowhere.org:custom/foo-1.0.0"
    result = deepcopy(original)
    assert result == original
    assert result._tag == original._tag


def test_tagged_string_copy():
    original = TaggedString("You're it!")
    original._tag = "tag:nowhere.org:custom/foo-1.0.0"
    result = copy(original)
    assert result == original
    assert result._tag == original._tag


def test_tagged_string_isinstance():
    value = TaggedString("You're it!")
    assert isinstance(value, str)


def test_tagged_string_base():
    value = TaggedString("You're it!")
    value._tag = "tag:nowhere.org:custom/foo-1.0.0"

    assert not (value == value.base)  # base is not a tagged  dict
    assert value.data == value.base  # but the data is

    assert isinstance(value.base, str)
    assert not isinstance(value.base, TaggedString)
    assert value.base.__class__ is str


ASDF_UNIT_TAG = "stsci.edu:asdf/unit/unit-1.0.0"
TAGGED_UNIT_URI = "asdf://stsci.edu/schemas/tagged_unit-1.0.0"
TAGGED_UNIT_SCHEMA = f"""
%YAML 1.1
---
$schema: http://stsci.edu/schemas/yaml-schema/draft-01
id: {TAGGED_UNIT_URI}

properties:
  unit:
    tag: {ASDF_UNIT_TAG}
    enum: [m, kg]
...
"""


def create_units():
    meter = TaggedString("m")
    meter._tag = ASDF_UNIT_TAG

    kilogram = TaggedString("kg")
    kilogram._tag = ASDF_UNIT_TAG

    return meter, kilogram


@pytest.mark.parametrize("unit", create_units())
def test_check_valid_str_enum(unit):
    """
    Regression test for issue #1254
        https://github.com/asdf-format/asdf/issues/1256

    This ensures that tagged strings can be properly validated against ``enum`` lists.
    """
    with asdf.config_context() as conf:
        conf.add_resource_mapping({TAGGED_UNIT_URI: TAGGED_UNIT_SCHEMA})
        schema = asdf.schema.load_schema(TAGGED_UNIT_URI)

        # This should not raise an exception (check_schema will raise error)
        asdf.schema.check_schema(schema)

        # This should not raise exception (validate will raise error)
        asdf.schema.validate({"unit": unit}, schema=schema)


def test_check_invalid_str_enum():
    """
    Ensure that a tagged string that is not in the ``enum`` list is properly handled.
    """
    with asdf.config_context() as conf:
        conf.add_resource_mapping({TAGGED_UNIT_URI: TAGGED_UNIT_SCHEMA})
        schema = asdf.schema.load_schema(TAGGED_UNIT_URI)

        # This should not raise an exception (check_schema will raise error)
        asdf.schema.check_schema(schema)

        unit = TaggedString("foo")
        unit._tag = ASDF_UNIT_TAG

        with pytest.raises(asdf.ValidationError, match=r"'foo' is not one of \['m', 'kg'\]"):
            asdf.schema.validate({"unit": unit}, schema=schema)


def _make_tagged_string(contents, tag):
    obj = TaggedString(contents)
    obj._tag = tag
    return obj


def _make_custom_tagged(tag):
    class CustomTagged(Tagged):
        pass

    obj = CustomTagged()
    obj._tag = tag
    return obj


def _make_custom_non_tagged(tag):
    class CustomNonTagged:
        pass

    obj = CustomNonTagged()
    obj._tag = tag
    return obj


@pytest.mark.parametrize(
    "obj, expected",
    (
        (TaggedList([0, 1, 2], "tag:nowhere.org:custom/foo-1.0.0"), "tag:nowhere.org:custom/foo-1.0.0"),
        (TaggedDict({"a": 1}, "tag:nowhere.org:custom/foo-1.0.0"), "tag:nowhere.org:custom/foo-1.0.0"),
        (_make_tagged_string("a", "tag:nowhere.org:custom/foo-1.0.0"), "tag:nowhere.org:custom/foo-1.0.0"),
        (_make_custom_tagged("tag:nowhere.org:custom/foo-1.0.0"), "tag:nowhere.org:custom/foo-1.0.0"),
        (_make_custom_non_tagged("tag:nowhere.org:custom/foo-1.0.0"), None),
    ),
)
def test_get_tag(obj, expected):
    """
    Test that get_tag returns the tag only for Tagged objects
    """
    assert get_tag(obj) == expected



from asdf import _compression as mcompression
from asdf.config import get_config


class Options:
    """
    Storage and compression options useful when reading or writing ASDF blocks.
    """

    def __init__(self, storage_type=None, compression_type=None, compression_kwargs=None, save_base=None):
        if storage_type is None:
            storage_type = get_config().all_array_storage or "internal"
        if save_base is None:
            save_base = get_config().default_array_save_base

        self._storage_type = None
        self._compression = None
        self._compression_kwargs = None

        # set via setters
        self.compression_kwargs = compression_kwargs
        self.compression = compression_type
        self.storage_type = storage_type
        self.save_base = save_base

    @property
    def storage_type(self):
        return self._storage_type

    @storage_type.setter
    def storage_type(self, storage_type):
        if storage_type not in ["internal", "external", "streamed", "inline"]:
            msg = "array_storage must be one of 'internal', 'external', 'streamed' or 'inline'"
            raise ValueError(msg)
        self._storage_type = storage_type

    @property
    def compression(self):
        return self._compression

    @compression.setter
    def compression(self, compression):
        msg = f"Invalid compression type: {compression}"
        if compression == "input":
            # "input" compression will validate as the ASDF compression module made
            # some assumptions about availability of information (that the input block
            # is known). The Options here do not have the same assumption.
            # For a block read from a file, it's options will be initialized with
            # the compression value read from the block header so we should never
            # see 'input' at this point.
            raise ValueError(msg)
        try:
            compression = mcompression.validate(compression)
        except ValueError:
            raise ValueError(msg)
        self._compression = compression

    @property
    def compression_kwargs(self):
        return self._compression_kwargs

    @compression_kwargs.setter
    def compression_kwargs(self, kwargs):
        if not kwargs:
            kwargs = {}
        self._compression_kwargs = kwargs

    @property
    def save_base(self):
        return self._save_base

    @save_base.setter
    def save_base(self, save_base):
        if not (isinstance(save_base, bool) or save_base is None):
            msg = "save_base must be a bool or None"
            raise ValueError(msg)
        self._save_base = save_base

    def __copy__(self):
        return type(self)(self._storage_type, self._compression, self._compression_kwargs)



from collections.abc import Mapping

import pytest

from asdf.resource import JsonschemaResourceMapping, ResourceManager, ResourceMappingProxy


def test_resource_manager():
    mapping1 = {
        "http://somewhere.org/schemas/foo-1.0.0": b"foo",
        "http://somewhere.org/schemas/bar-1.0.0": b"bar",
    }
    mapping2 = {
        "http://somewhere.org/schemas/foo-1.0.0": b"duplicate foo",
        "http://somewhere.org/schemas/baz-1.0.0": b"baz",
        "http://somewhere.org/schemas/foz-1.0.0": "foz",
    }
    manager = ResourceManager([mapping1, mapping2])

    assert isinstance(manager, Mapping)

    assert len(manager) == 4
    assert set(manager) == {
        "http://somewhere.org/schemas/foo-1.0.0",
        "http://somewhere.org/schemas/bar-1.0.0",
        "http://somewhere.org/schemas/baz-1.0.0",
        "http://somewhere.org/schemas/foz-1.0.0",
    }
    assert "http://somewhere.org/schemas/foo-1.0.0" in manager
    assert manager["http://somewhere.org/schemas/foo-1.0.0"] == b"foo"
    assert "http://somewhere.org/schemas/bar-1.0.0" in manager
    assert manager["http://somewhere.org/schemas/bar-1.0.0"] == b"bar"
    assert "http://somewhere.org/schemas/baz-1.0.0" in manager
    assert manager["http://somewhere.org/schemas/baz-1.0.0"] == b"baz"
    assert "http://somewhere.org/schemas/foz-1.0.0" in manager
    assert manager["http://somewhere.org/schemas/foz-1.0.0"] == b"foz"

    with pytest.raises(KeyError, match=r"http://somewhere.org/schemas/missing-1.0.0"):
        manager["http://somewhere.org/schemas/missing-1.0.0"]

    # Confirm that the repr string is reasonable:
    assert "len: 4" in repr(manager)


def test_jsonschema_resource_mapping():
    mapping = JsonschemaResourceMapping()
    assert isinstance(mapping, Mapping)

    assert len(mapping) == 1
    assert set(mapping) == {"http://json-schema.org/draft-04/schema"}
    assert "http://json-schema.org/draft-04/schema" in mapping
    assert b"http://json-schema.org/draft-04/schema" in mapping["http://json-schema.org/draft-04/schema"]

    assert repr(mapping) == "JsonschemaResourceMapping()"


def test_proxy_is_mapping():
    assert isinstance(ResourceMappingProxy({}), Mapping)


def test_proxy_maybe_wrap():
    mapping = {
        "http://somewhere.org/resources/foo": "foo",
        "http://somewhere.org/resources/bar": "bar",
    }

    proxy = ResourceMappingProxy.maybe_wrap(mapping)
    assert proxy.delegate is mapping
    assert ResourceMappingProxy.maybe_wrap(proxy) is proxy

    with pytest.raises(TypeError, match=r"Resource mapping must implement the Mapping interface"):
        ResourceMappingProxy.maybe_wrap([])


def test_proxy_properties():
    mapping = {
        "http://somewhere.org/resources/foo": "foo",
        "http://somewhere.org/resources/bar": "bar",
    }

    proxy = ResourceMappingProxy(mapping, package_name="foo", package_version="1.2.3")

    assert len(proxy) == len(mapping)
    assert set(proxy.keys()) == set(mapping.keys())
    for uri in mapping:
        assert proxy[uri] is mapping[uri]

    assert proxy.package_name == "foo"
    assert proxy.package_version == "1.2.3"
    assert proxy.class_name.endswith("dict")


def test_proxy_hash_and_eq():
    mapping = {
        "http://somewhere.org/resources/foo": "foo",
        "http://somewhere.org/resources/bar": "bar",
    }
    proxy1 = ResourceMappingProxy(mapping)
    proxy2 = ResourceMappingProxy(mapping, package_name="foo", package_version="1.2.3")

    assert proxy1 == proxy2
    assert hash(proxy1) == hash(proxy2)
    assert proxy1 != mapping
    assert proxy2 != mapping


def test_proxy_repr():
    mapping = {
        "http://somewhere.org/resources/foo": "foo",
        "http://somewhere.org/resources/bar": "bar",
    }

    proxy = ResourceMappingProxy(mapping, package_name="foo", package_version="1.2.3")

    assert ".dict" in repr(proxy)
    assert "package: foo==1.2.3" in repr(proxy)
    assert "len: 2" in repr(proxy)

    empty_proxy = ResourceMappingProxy({})

    assert ".dict" in repr(empty_proxy)
    assert "package: (none)" in repr(empty_proxy)
    assert "len: 0" in repr(empty_proxy)



from asdf import treeutil


def test_get_children():
    parent = ["foo", "bar"]
    assert treeutil.get_children(parent) == [(0, "foo"), (1, "bar")]

    parent = ("foo", "bar")
    assert treeutil.get_children(parent) == [(0, "foo"), (1, "bar")]

    parent = {"foo": "bar", "ding": "dong"}
    assert sorted(treeutil.get_children(parent)) == sorted([("foo", "bar"), ("ding", "dong")])

    parent = "foo"
    assert treeutil.get_children(parent) == []

    parent = None
    assert treeutil.get_children(parent) == []


def test_is_container():
    for value in [[], {}, ()]:
        assert treeutil.is_container(value) is True

    for value in ["foo", 12, 13.9827]:
        assert treeutil.is_container(value) is False


def test_walk_and_modify_shared_references():
    target = {"foo": "bar"}
    nested_in_dict = {"target": target}
    nested_in_list = [target]
    tree = {"target": target, "nested_in_dict": nested_in_dict, "nested_in_list": nested_in_list}

    assert tree["target"] is tree["nested_in_dict"]["target"]
    assert tree["target"] is tree["nested_in_list"][0]

    def _callback(node):
        if "foo" in node:
            return {"foo": "baz"}

        return node

    result = treeutil.walk_and_modify(tree, _callback)

    assert result is not tree
    assert result["target"] is not target
    assert result["target"]["foo"] == "baz"
    assert result["target"] is result["nested_in_dict"]["target"]
    assert result["target"] is result["nested_in_list"][0]



"""
Implementation of command for displaying available tags in asdf
"""

import sys

from asdf import AsdfFile

from .main import Command

__all__ = ["list_tags"]


class TagLister(Command):  # pragma: no cover
    """This class is the plugin implementation for the asdftool runner."""

    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "tags",
            help="List currently available tags",
            description="""Lists currently available tags.""",
        )

        parser.add_argument(
            "-d",
            "--display-classes",
            action="store_true",
            help="""Display associated class names in addition to tags""",
        )

        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        return list_tags(display_classes=args.display_classes)


def _format_type(typ):
    if isinstance(typ, str):
        return typ

    return f"{typ.__module__}.{typ.__name__}"


def list_tags(display_classes=False, iostream=sys.stdout):
    """Function to list tags"""
    af = AsdfFile()

    tag_pairs = []
    for tag in af.extension_manager._converters_by_tag:
        tag_pairs.append((tag, af.extension_manager.get_converter_for_tag(tag).types))

    for tag, types in sorted(tag_pairs, key=lambda pair: pair[0]):
        string = str(tag)
        if display_classes:
            string += ":  " + ", ".join(_format_type(t) for t in types)
        iostream.write(string + "\n")



import warnings

import numpy as np

import asdf

__all__ = [
    "assert_tree_match",
]


def assert_tree_match(old_tree, new_tree):
    """
    Assert that two ASDF trees match.

    Parameters
    ----------
    old_tree : ASDF tree

    new_tree : ASDF tree
    """

    seen = set()
    ignore_keys = {"asdf_library", "history"}

    def recurse(old, new):
        if id(old) in seen or id(new) in seen:
            return

        seen.add(id(old))
        seen.add(id(new))

        if isinstance(old, dict) and isinstance(new, dict):
            assert {x for x in old if x not in ignore_keys} == {x for x in new if x not in ignore_keys}
            for key in old:
                if key not in ignore_keys:
                    recurse(old[key], new[key])
        elif isinstance(old, (list, tuple)) and isinstance(new, (list, tuple)):
            assert len(old) == len(new)
            for a, b in zip(old, new):
                recurse(a, b)
        elif all([isinstance(obj, (np.ndarray, asdf.tags.core.NDArrayType)) for obj in (old, new)]):
            with warnings.catch_warnings():
                # The oldest deps job tests against versions of numpy where this
                # testing function raised a FutureWarning but still functioned
                # as expected
                warnings.filterwarnings("ignore", category=FutureWarning)
                if old.dtype.fields:
                    if not new.dtype.fields:
                        msg = "arrays not equal"
                        raise AssertionError(msg)
                    for f in old.dtype.fields:
                        np.testing.assert_array_equal(old[f], new[f])
                else:
                    np.testing.assert_array_equal(old.__array__(), new.__array__())
        else:
            assert old == new

    recurse(old_tree, new_tree)



from asdf.extension import Converter


class ReferenceConverter(Converter):
    tags = []
    types = ["asdf.reference.Reference"]

    def to_yaml_tree(self, obj, tag, ctx):
        from asdf.generic_io import relative_uri

        base_uri = None
        if ctx._blocks._write_fd is not None and ctx._blocks._write_fd.uri is not None:
            base_uri = ctx._blocks._write_fd.uri
        elif ctx.url is not None:
            base_uri = ctx.url
        uri = relative_uri(base_uri, obj._uri) if base_uri is not None else obj._uri
        return {"$ref": uri}

    def from_yaml_tree(self, node, tag, ctx):
        raise NotImplementedError()

    def select_tag(self, obj, tags, ctx):
        return None



import io
import os

import pytest

import asdf
from asdf import generic_io


def test_no_yaml_end_marker(tmp_path):
    content = b"""#ASDF 1.0.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.0.0
foo: bar...baz
baz: 42
    """
    path = os.path.join(str(tmp_path), "test.asdf")

    buff = io.BytesIO(content)
    with pytest.raises(ValueError, match=r"End of YAML marker not found"), asdf.open(buff):
        pass

    buff.seek(0)
    with pytest.raises(ValueError, match=r"End of YAML marker not found"), asdf.open(buff):
        pass

    with open(path, "wb") as fd:
        fd.write(content)

    with open(path, "rb") as fd, pytest.raises(ValueError, match=r"End of YAML marker not found"), asdf.open(fd):
        pass


def test_no_final_newline(tmp_path):
    content = b"""#ASDF 1.0.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.0.0
foo: ...bar...
baz: 42
..."""
    path = os.path.join(str(tmp_path), "test.asdf")

    buff = io.BytesIO(content)
    with asdf.open(buff) as ff:
        assert len(ff.tree) == 2

    buff.seek(0)
    fd = generic_io.InputStream(buff, "r")
    with asdf.open(fd) as ff:
        assert len(ff.tree) == 2

    with open(path, "wb") as fd:
        fd.write(content)

    with open(path, "rb") as fd, asdf.open(fd) as ff:
        assert len(ff.tree) == 2


def test_no_asdf_header(tmp_path):
    content = b"What? This ain't no ASDF file"

    path = os.path.join(str(tmp_path), "test.asdf")

    buff = io.BytesIO(content)
    with pytest.raises(
        ValueError,
        match=r"Does not appear to be a ASDF file.",
    ):
        asdf.open(buff)

    with open(path, "wb") as fd:
        fd.write(content)

    with (
        open(path, "rb") as fd,
        pytest.raises(
            ValueError,
            match=r"Does not appear to be a ASDF file.",
        ),
    ):
        asdf.open(fd)


def test_empty_file():
    buff = io.BytesIO(b"#ASDF 1.0.0\n")
    buff.seek(0)

    with asdf.open(buff) as ff:
        assert ff.tree == {}
        assert len(ff._blocks.blocks) == 0

    buff = io.BytesIO(b"#ASDF 1.0.0\n#ASDF_STANDARD 1.0.0")
    buff.seek(0)

    with asdf.open(buff) as ff:
        assert ff.tree == {}
        assert len(ff._blocks.blocks) == 0


def test_not_asdf_file():
    buff = io.BytesIO(b"SIMPLE")
    buff.seek(0)

    with (
        pytest.raises(
            ValueError,
            match=r"Does not appear to be a ASDF file.",
        ),
        asdf.open(buff),
    ):
        pass

    buff = io.BytesIO(b"SIMPLE\n")
    buff.seek(0)

    with (
        pytest.raises(
            ValueError,
            match=r"Does not appear to be a ASDF file.",
        ),
        asdf.open(buff),
    ):
        pass


def test_junk_file():
    buff = io.BytesIO(b"#ASDF 1.0.0\nFOO")
    buff.seek(0)

    with pytest.raises(ValueError, match=r"Invalid content between header and tree"), asdf.open(buff):
        pass


def test_invalid_header_version():
    buff = io.BytesIO(b"#ASDF foo\n")
    buff.seek(0)

    with pytest.raises(ValueError, match=r"Unparsable version in ASDF file"), asdf.open(buff):
        pass


def test_block_mismatch():
    # This is a file with a single small block, followed by something
    # that has an invalid block magic number.

    buff = io.BytesIO(
        b"#ASDF 1.0.0\n\xd3BLK\x00\x28\0\0\0\0\0\0\0\x01\0\0\0\0\0\0\0\x01"
        b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0FOOBAR",
    )

    buff.seek(0)
    with pytest.raises(ValueError, match=r"Header size must be >= 48"), asdf.open(buff):
        pass


def test_block_header_too_small():
    # The block header size must be at least 40

    buff = io.BytesIO(b"#ASDF 1.0.0\n\xd3BLK\0\0")

    buff.seek(0)
    with pytest.raises(ValueError, match=r"Header size must be >= 48"), asdf.open(buff):
        pass


def test_invalid_version(tmp_path):
    content = b"""#ASDF 0.1.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-0.1.0
foo : bar
..."""
    buff = io.BytesIO(content)
    with pytest.raises(ValueError, match=r"Unsupported ASDF file format version*"), asdf.open(buff):
        pass


def test_valid_version(tmp_path):
    content = b"""#ASDF 1.0.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.0.0
foo : bar
..."""
    buff = io.BytesIO(content)
    with asdf.open(buff) as ff:
        version = ff.file_format_version

    assert version.major == 1
    assert version.minor == 0
    assert version.patch == 0



from contextlib import nullcontext

import pytest

import asdf
from asdf.exceptions import ValidationError

# test cases with:
# - custom schema path
# - tree
# - expected error (or None)
TEST_CASES = [
    (
        "custom_schema.yaml",
        {"stuff": 42, "other_stuff": "hello"},
        ".* is a required property",
    ),
    (
        "custom_schema.yaml",
        {"foo": {"x": 42, "y": 10}, "bar": {"a": "hello", "b": "banjo"}},
        None,
    ),
    (
        "custom_schema_definitions.yaml",
        {"forb": {"biz": "hello", "baz": "world"}},
        ".* is a required property",
    ),
    (
        "custom_schema_definitions.yaml",
        {"thing": {"biz": "hello", "baz": "world"}},
        None,
    ),
    (
        "custom_schema_external_ref.yaml",
        {"foo": asdf.tags.core.Software(name="Microsoft Windows", version="95")},
        None,
    ),
    (
        "custom_schema_external_ref.yaml",
        {"foo": False},
        "False is not valid under any of the given schemas",
    ),
]


@pytest.fixture(params=[lambda x: x, str])
def as_pathlib(request):
    """
    Fixture to test both pathlib.Path and str.
    """
    return request.param


@pytest.fixture
def schema_name(request, test_data_path, as_pathlib):
    """
    Fixture to convert the provided schema name to a path.
    """
    return as_pathlib(test_data_path / request.param)


@pytest.mark.parametrize(
    "schema_name, tree, expected_error",
    TEST_CASES,
    indirect=["schema_name"],
)
def test_custom_validation_write_to(tmp_path, schema_name, tree, expected_error):
    asdf_file = tmp_path / "out.asdf"

    ctx = pytest.raises(ValidationError, match=expected_error) if expected_error else nullcontext()
    with ctx:
        asdf.AsdfFile(tree, custom_schema=schema_name).write_to(asdf_file)


@pytest.mark.parametrize(
    "schema_name, tree, expected_error",
    TEST_CASES,
    indirect=["schema_name"],
)
def test_custom_validation_open(tmp_path, schema_name, tree, expected_error):
    asdf_file = tmp_path / "out.asdf"

    asdf.AsdfFile(tree).write_to(asdf_file)
    ctx = pytest.raises(ValidationError, match=expected_error) if expected_error else nullcontext()

    with ctx, asdf.open(asdf_file, custom_schema=schema_name):
        pass


@pytest.mark.parametrize(
    "schema_name, tree, expected_error",
    TEST_CASES,
    indirect=["schema_name"],
)
def test_custom_validation_validate(schema_name, tree, expected_error):

    ctx = pytest.raises(ValidationError, match=expected_error) if expected_error else nullcontext()

    with ctx:
        asdf.AsdfFile(tree, custom_schema=schema_name).validate()


@pytest.mark.parametrize(
    "schema_name, tree, expected_error",
    TEST_CASES,
    indirect=["schema_name"],
)
def test_custom_validation_dumps(schema_name, tree, expected_error):

    ctx = pytest.raises(ValidationError, match=expected_error) if expected_error else nullcontext()

    with ctx:
        asdf.dumps(tree, custom_schema=schema_name)


@pytest.mark.parametrize(
    "schema_name, tree, expected_error",
    TEST_CASES,
    indirect=["schema_name"],
)
def test_custom_validation_loads(schema_name, tree, expected_error):

    contents = asdf.dumps(tree)

    ctx = pytest.raises(ValidationError, match=expected_error) if expected_error else nullcontext()

    with ctx:
        asdf.loads(contents, custom_schema=schema_name)


@pytest.mark.parametrize(
    "schema_name, tree, expected_error",
    TEST_CASES,
    indirect=["schema_name"],
)
def test_custom_validation_dump(tmp_path, schema_name, tree, expected_error):
    asdf_file = tmp_path / "out.asdf"

    ctx = pytest.raises(ValidationError, match=expected_error) if expected_error else nullcontext()

    with ctx:
        asdf.dump(tree, asdf_file, custom_schema=schema_name)


@pytest.mark.parametrize(
    "schema_name, tree, expected_error",
    TEST_CASES,
    indirect=["schema_name"],
)
def test_custom_validation_load(tmp_path, schema_name, tree, expected_error):
    asdf_file = tmp_path / "out.asdf"

    asdf.AsdfFile(tree).write_to(asdf_file)
    ctx = pytest.raises(ValidationError, match=expected_error) if expected_error else nullcontext()

    with ctx:
        asdf.load(asdf_file, custom_schema=schema_name)



"""
Test that warn_on_failed_conversion works as expected for:
    - lazy and non-lazy trees
    - nested objects
    - generator producing (yielding) converters

Broken out into a regtest since there is significant
test setup.

https://github.com/asdf-format/asdf/issues/1948
"""

import pytest

import asdf

test_dict_tag_uri = "asdf://somewhere.org/tags/test_dict-1.0.0"
failing_dict_tag_uri = "asdf://somewhere.org/tags/failing_dict-1.0.0"
failing_yield_dict_tag_uri = "asdf://somewhere.org/tags/failing_yield_dict-1.0.0"


class MyDict:
    def __init__(self, data):
        self.data = data


class FailingDict(MyDict):
    pass


class FailingYieldDict(MyDict):
    pass


class MyDictConverter:
    tags = [test_dict_tag_uri]
    types = [MyDict]
    lazy = True

    def to_yaml_tree(self, obj, tag, ctx):
        return obj.data

    def from_yaml_tree(self, node, tag, ctx):
        return MyDict(node)


class FailingDictConverter:
    tags = [failing_dict_tag_uri]
    types = [FailingDict]
    lazy = True

    def to_yaml_tree(self, obj, tag, ctx):
        return obj.data

    def from_yaml_tree(self, node, tag, ctx):
        # always make reading a failing_dict tag fail
        raise Exception("FailingDict failed")


class FailingYieldDictConverter:
    tags = [failing_yield_dict_tag_uri]
    types = [FailingYieldDict]
    lazy = True

    def to_yaml_tree(self, obj, tag, ctx):
        return obj.data

    def from_yaml_tree(self, node, tag, ctx):
        raise Exception("FailingYieldDict failed")
        yield {}


class TestExtension:
    extension_uri = "asdf://somewhere.org/extensions/minimum-1.0.0"
    converters = [MyDictConverter(), FailingDictConverter(), FailingYieldDictConverter()]
    tags = [test_dict_tag_uri, failing_dict_tag_uri, failing_yield_dict_tag_uri]


@pytest.fixture()
def enable_test_extension():
    with asdf.config_context() as cfg:
        cfg.add_extension(TestExtension())
        yield


def test_failed_conversion_warns(enable_test_extension):
    tree = {
        "test_dict": MyDict({"failing_dict": FailingDict({"a": 1})}),
        "failing_yield_dict": FailingYieldDict({"b": 2}),
    }
    asdf_str = asdf.dumps(tree)
    with asdf.config_context() as cfg:
        cfg.warn_on_failed_conversion = True
        with (
            pytest.warns(asdf.exceptions.AsdfConversionWarning, match="FailingDict failed"),
            pytest.warns(asdf.exceptions.AsdfConversionWarning, match="FailingYieldDict failed"),
        ):
            read_obj = asdf.loads(asdf_str)
        test_dict = read_obj["test_dict"]
        assert isinstance(test_dict, MyDict)
        failing_dict = read_obj["test_dict"].data["failing_dict"]
        assert isinstance(failing_dict, asdf.tagged.TaggedDict)
        assert failing_dict._tag == failing_dict_tag_uri
        assert failing_dict["a"] == 1
        failing_yield_dict = read_obj["failing_yield_dict"]
        assert isinstance(failing_yield_dict, asdf.tagged.TaggedDict)
        assert failing_yield_dict._tag == failing_yield_dict_tag_uri
        assert failing_yield_dict["b"] == 2


def test_lazy_failed_conversion_warns(enable_test_extension, tmp_path):
    test_path = tmp_path / "test.asdf"
    tree = {
        "test_dict": MyDict({"failing_dict": FailingDict({"a": 1})}),
        "failing_yield_dict": FailingYieldDict({"b": 2}),
    }
    asdf.dump(tree, test_path)
    with asdf.config_context() as cfg:
        cfg.warn_on_failed_conversion = True
        cfg.lazy_tree = True
        with asdf.open(test_path) as af:
            test_dict = af["test_dict"]
            assert isinstance(test_dict, MyDict)
            with pytest.warns(asdf.exceptions.AsdfConversionWarning, match="FailingDict failed"):
                failing_dict = af["test_dict"].data["failing_dict"]
            with pytest.warns(asdf.exceptions.AsdfConversionWarning, match="FailingYieldDict failed"):
                failing_yield_dict = af["failing_yield_dict"]
            assert isinstance(failing_dict, asdf.tagged.TaggedDict)
            assert failing_dict._tag == failing_dict_tag_uri
            assert failing_dict["a"] == 1
            assert isinstance(failing_yield_dict, asdf.tagged.TaggedDict)
            assert failing_yield_dict._tag == failing_yield_dict_tag_uri
            assert failing_yield_dict["b"] == 2



"""
Implementation of the asdf.info(...) function.  This is just a thin wrapper
around _display module code.
"""

from contextlib import contextmanager
from pathlib import Path

from ._asdf import AsdfFile, open_asdf
from ._display import DEFAULT_MAX_COLS, DEFAULT_MAX_ROWS, DEFAULT_SHOW_VALUES

__all__ = ["info"]


def info(
    node_or_path,
    max_rows=DEFAULT_MAX_ROWS,
    max_cols=DEFAULT_MAX_COLS,
    show_values=DEFAULT_SHOW_VALUES,
    show_blocks=False,
):
    """
    Print a rendering of an ASDF tree or sub-tree to stdout.

    Parameters
    ----------
    node_or_path : str, pathlib.Path, asdf.AsdfFile, or any ASDF tree node
        The tree or sub-tree to render.  Strings and Path objects
        will first be passed to asdf.open(...).

    max_rows : int, tuple, or None, optional
        Maximum number of lines to print.  Nodes that cannot be
        displayed will be elided with a message.
        If int, constrain total number of displayed lines.
        If tuple, constrain lines per node at the depth corresponding \
            to the tuple index.
        If None, display all lines.

    max_cols : int or None, optional
        Maximum length of line to print.  Nodes that cannot
        be fully displayed will be truncated with a message.
        If int, constrain length of displayed lines.
        If None, line length is unconstrained.

    show_values : bool, optional
        Set to False to disable display of primitive values in
        the rendered tree.

    show_blocks: bool, optional
        Set to True to also print a table of block header fields
        for each block in the file.
    """
    with _manage_node(node_or_path) as node:
        node.info(max_rows=max_rows, max_cols=max_cols, show_values=show_values, show_blocks=show_blocks)


@contextmanager
def _manage_node(node_or_path):
    if isinstance(node_or_path, (str, Path)):
        with open_asdf(node_or_path) as af:
            yield af

    elif isinstance(node_or_path, AsdfFile):
        yield node_or_path

    else:
        yield AsdfFile(node_or_path)
